'''
Record packets from the current run and compute various statistics.

'''

import time
import logging
from collections import deque, defaultdict
import json
import argparse

import requests
from xylem import Consumer, protocol
from xylem.EventHandler import EventHandler
from larpix.larpix import Packet
from larpixgeometry import layouts

import larpixdaq.packetformat as pformat

class OnlineMonitor(object):
    """Record packets from the current run and compute various statistics.

    Run data provides the online data monitor for the LArPix DAQ. It
    tracks the packet rate and can send packets for manual inspection.
    It also tracks each pixel's recent and max data rate as well as an
    ADC histogram.

    :param core_address: the full TCP address (including port number)
        that data will be published to
    """

    def __init__(self, core_address):
        consumer_args = {
                'core_address': core_address,
                'heartbeat_time_ms': 300,
                'action_docs': {
                    'data_rate': '''data_rate()
                        Return the average data rate for the packets
                        received so far.''',
                    'packets': '''packets()
                        Return the packets as a bytestream with a 2-byte
                        delimiter of 0xAAAA.''',
                    'messages': '''messages()
                        Return the messages.''',
                }

        }
        self._consumer = Consumer(name='Online monitor', connections=['AGGREGATOR'],
                **consumer_args)
        self._consumer.register_action('retrieve_pixel_layout',
                self.retrieve_pixel_layout, self.retrieve_pixel_layout.__doc__)
        self._consumer.register_action('load_pixel_layout',
                self.load_pixel_layout, self.load_pixel_layout.__doc__)
        self._consumer.addHandler(EventHandler('data_message',
            self.handle_new_data))
        self._consumer.addHandler(EventHandler('data_message',
            self.maybe_send_update))
        self._consumer.addHandler(EventHandler('info_message',
            self.handle_new_message))
        self._consumer.addHandler(EventHandler('info_message',
            self.send_message_update))
        self._use_requests = True
        self.packets = deque([], 100000)
        self.timestamps = defaultdict(int)
        self.pixel_rates=defaultdict(([0]*832).copy)
        self.max_pixel_rates = [0]*832
        self.messages = []
        self.datarates = deque([], 100)
        self.datarate_timestamps = deque([], 100)
        self.adcs = deque([], 1000)
        self.start_time = 0
        self._sent_index = 0
        self.runno = 0
        self.state = self._consumer.state
        self.layout = {'chips':[], 'pixels':[]}
        self.pixel_lookup = {}
        self.last_second = int(time.time())
        return

    def handle_new_message(self, origin, header, message):
        """Store new info messages."""
        self.messages.append(message)
        if (header['component'] == 'LArPix board' and message ==
                'Beginning run'):
            self._consumer.log('INFO', 'Received start message')
            self._start_run()
        elif (header['component'] == 'LArPix board' and message ==
                'Ending run'):
            self._consumer.log('INFO', 'Received end message')
            self._end_run()

    def handle_new_data(self, origin, header, data):
        """Store new data packets and save data rate and ADCs."""
        packets = pformat.fromBytes(data)
        self.packets.extend(packets)
        now = int(time.time())
        self.timestamps[now] += len(packets)
        pixel_rates_now = self.pixel_rates[now]
        for packet in packets:
            if packet.packet_type == Packet.DATA_PACKET:
                chip_key = str(packet.chip_key)
                channelid = packet.channel_id
                pixel_list = self.pixel_lookup.get(chip_key, None)
                if pixel_list is not None:
                    pixelid = pixel_list[channelid]
                    if pixelid is not None:
                        pixel_rates_now[pixelid] += 1
                        if (pixel_rates_now[pixelid] >
                                self.max_pixel_rates[pixelid]):
                            self.max_pixel_rates[pixelid] = (
                                    pixel_rates_now[pixelid])
        self.adcs.extend(p.dataword for p in packets
                if p.packet_type == Packet.DATA_PACKET)

    def maybe_send_update(self, *args):
        """Send an update to the webserver once per second.

        :param args: ignored
        """
        if not self._use_requests:
            return
        now = int(time.time())
        next_tick = now != self.last_second
        if next_tick:
            self.datarates.append(self.timestamps[self.last_second])
            self.datarate_timestamps.append(self.last_second)
            pixel_rates_last_second = self.pixel_rates[self.last_second][:]
            del self.timestamps[self.last_second]
            del self.pixel_rates[self.last_second]
            self.last_second = now
            try:
                r = requests.post('http://localhost:5000/packets',
                        json={'rate':self._data_rate(),
                            'packets':self._packets()[-100:][::-1],
                            'messages':self._messages()[-100:][::-1],
                            'rate_list':list(self.datarates),
                            'rate_times':list(self.datarate_timestamps),
                            'adcs': list(self.adcs),
                            'rate_bypixel': pixel_rates_last_second,
                            'maxrate_bypixel': self.max_pixel_rates,
                            })
            except requests.ConnectionError as e:
                self._use_requests = False
                self._consumer.log('WARNING', 'Failed to send packets '
                        'to server: %s; turning off web requests' % e)

    def send_message_update(self, *args):
        """Send an update containing info messages.

        :param args: ignored
        """
        if not self._use_requests:
            return
        try:
            r = requests.post('http://localhost:5000/packets',
                    json={'messages':self._messages()[-100:][::-1],}
                    )
        except requests.ConnectionError as e:
            self._use_requests = False
            self._consumer.log('WARNING', 'Failed to send packets '
                    'to server: %s; turning off web requests' % e)

    def create_pixel_lookup(self, chip_pixel_list):
        """Create a pixel lookup from a given list of chip-pixel
        assignments.

        chip_pixel_list is of the form

        ```
        [
          [chip0id, [ch0pixel, ch1pixel, ...]],
          [chip1id, [ch0pixel, ch1pixel, ...]],
          ...
        ]
        ```
        """
        pixel_lookup = {}
        for (chipid, pixels) in chip_pixel_list:
            pixel_lookup[chipid] = pixels
        return pixel_lookup

    def create_chip_lookup(self, chip_pixel_list):
        """Create a chip+channel lookup based on pixel ID.

        The input chip_pixel_list is of the form

        ```
        [
          [chip0id, [ch0pixel, ch1pixel, ...]],
          [chip1id, [ch0pixel, ch1pixel, ...]],
          ...
        ]
        ```

        The output is of the form

        ```
        {
          pixel0id: {'channel': pixel0channel, 'chip': pixel0chip},
          pixel1id: {'channel': pixel1channel, 'chip': pixel1chip},
          ...
        }
        ```
        """
        chip_lookup = {}
        for (chipid, pixels) in chip_pixel_list:
            for (channelid, pixelid) in enumerate(pixels):
                chip_lookup[pixelid] = {
                        'channel': channelid,
                        'chip': chipid,
                        }
        return chip_lookup

    def retrieve_pixel_layout(self):
        '''
        retrieve_pixel_layout()

        Return the current pixel layout and pixel->{chip, channel}
        mapping.

        '''
        return {
                'layout': self.layout,
                'lookup': self.chip_lookup,
                }

    def load_pixel_layout(self, name):
        '''
        load_pixel_layout(name)

        Retrieve and store the pixel layout from larpix-geometry.

        '''
        self.layout = layouts.load(name)
        for entry in self.layout['chips']:
            entry[0] = '1-1-%d' % entry[0]
        self.pixel_lookup = self.create_pixel_lookup(self.layout['chips'])
        self.chip_lookup = self.create_chip_lookup(self.layout['chips'])
        return {
                'layout': self.layout,
                'lookup': self.chip_lookup,
                }

    def _data_rate(self):
        '''
        Return the average data rate for the packets received so far.

        '''
        npackets = len(self.packets)
        time_elapsed = time.time() - self.start_time
        return '%.2f' % (npackets/time_elapsed)

    def _packets(self):
        '''
        Return a bytestream of all the packets received, converted to a
        string in base64 encoding.

        '''
        return pformat.toDict(self.packets)

    def _messages(self):
        '''
        Return the messages.

        '''
        return self.messages

    def _prepare_run(self):
        self.packets.clear()
        self.datarates.clear()
        self.datarate_timestamps.clear()
        self.pixel_rates.clear()
        self.max_pixel_rates = [0] * 832
        self.adcs.clear()

    def _start_run(self):
        self.runno += 1
        self.start_time = time.time()

    def _end_run(self):
        pass

    def run(self):
        t_last_send = time.time()
        try:
            r = requests.post('http://localhost:5000/packets', json={'rate':0,
                'packets':[]})
        except requests.ConnectionError as e:
            self._use_requests = False
            self._consumer.log('WARNING', 'Failed to send packets '
                    'to server: %s; turning off web requests' % e)
        last_second = int(time.time())
        while True:
            messages = self._consumer.receive(1)
            if self.state != self._consumer.state:
                old_state = self.state
                new_state = self._consumer.state
                if new_state == 'READY':
                    self._prepare_run()
                self.state = self._consumer.state

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Launch the data '
            'consumer providing the online data monitor')
    parser.add_argument('--core', default='tcp://127.0.0.1',
            help='The address of the DAQ Core, not including port number')
    args = parser.parse_args()
    monitor = OnlineMonitor(args.core + ':5551')
    try:
        monitor.run()
    except KeyboardInterrupt:
        pass
    finally:
        monitor._consumer.cleanup()



from larpix.logger import Logger
from larpixdaq.packetformat import toBytes

class DAQLogger(Logger):
    '''
    Logs larpix-control packets to the DAQ pipeline.

    '''
    def __init__(self, producer):
        super(DAQLogger, self).__init__(self)
        self.producer = producer

    def record(self, data, direction=Logger.WRITE):
        if self.is_enabled():
            to_produce = toBytes(data)
            self.producer.produce(to_produce)
        else:
            pass



from larpixdaq.routines import Routine
from larpix.quickstart import quickcontroller

def _quickstart(board, send_data, send_info, board_name):
    '''
    quickstart(board_name)

    Start up the board and configure ASICs to a quiescent state.

    '''
    send_info('Running quickstart')
    new_board = quickcontroller(board_name, io=board.io,
            logger=board.logger)
    send_info('Completed quickstart')
    result = 'success'
    return new_board, result

quickstart = Routine('quickstart', _quickstart, ['board'])

registration = {
        'quickstart': quickstart,
        }



import argparse

from xylem import Consumer
from xylem.EventHandler import EventHandler
from larpix.logger.h5_logger import HDF5Logger

from larpixdaq.packetformat import fromBytes

class OfflineStorage(object):
    """Record all received packets in offline storage.

    The offline storage script stores LArPix data to disk using the
    LArPix+HDF5 file format.

    :var consumer: the xylem Consumer object used to receive data
    :var state: the DAQ State of the xylem Consumer component
    :var logger: the LArPix Logger object used to save data to disk. Can
        be ``None`` if current state is not READY or RUN.

    :param core_address: the full TCP address (including port number) of
        the DAQ core
    """

    def __init__(self, core_address):
        consumer_args = {
                'core_address': core_address,
                'heartbeat_time_ms': 300,
        }
        self.consumer = Consumer(name='Offline storage',
                connections=['AGGREGATOR'], **consumer_args)
        self.state = ''
        self.consumer.addHandler(EventHandler('data_message',
            self.handle_new_data))
        self.logger = None

    def handle_new_data(self, origin, header, data):
        """Save new data to disk.

        Parameters are defined by the ``xylem.EventHandler`` interface.
        """
        if ((self.state == 'RUN' or self.state == 'READY')
                and self.logger is not None):
            packets = fromBytes(data)
            self.logger.record(packets)
        else:
            return

    def run(self):
        """Initiate the event loop of reading and saving data."""
        try:
            while True:
                messages = self.consumer.receive(None)
                if self.state != self.consumer.state:
                    old_state = self.state
                    new_state = self.consumer.state
                    if old_state == 'RUN':
                        if self.logger is not None:
                            self.logger.flush()
                            self.logger.disable()
                            self.logger = None
                    if new_state == 'READY':
                        self.logger = HDF5Logger(directory='runs')
                        self.logger.enable()
                        self.consumer.log('INFO', 'Storing data in file'
                                ' %s' % self.logger.filename)
                    self.state = new_state
        finally:
            if self.logger is not None:
                self.logger.flush()
                self.logger.disable()

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Launch the data '
            'consumer to save LArPix data to disk')
    parser.add_argument('--core', default='tcp://127.0.0.1',
            help='The address of the DAQ Core, not including port number')
    args = parser.parse_args()
    offline_storage = OfflineStorage(args.core + ':5551')
    try:
        offline_storage.run()
    except KeyboardInterrupt:
        pass
    finally:
        offline_storage.consumer.cleanup()



"""Defines routines for LArPix DAQ components.

Routines allow a DAQ operator to execute a predefined sequence of
operations. Routines are Python functions wrapped in a ``Routine``
object that stores metadata such as documentation and argument lists.

Using routines
--------------

Existing routines can be loaded by importing this module and calling
``init_routines()``. The optional argument can provide a custom location
for routine files. The default is ``larpixdaq/routines``, i.e. within
this Python package. ``init_routines`` modifies the global ``ROUTINES``
dict, which maps routine names to ``Routine`` objects. The routine's
function can be accessed at ``ROUTINES[name].func``, and called with the
appropriate parameters.

Routines all take 3 arguments as part of the routines interface, plus
any other positional arguments that are part of the routine's
implementation. The 3 required arguments allow the routine to interact
with the DAQ system. See the documentation for the ``Routine`` class for
details.

Routines all return a 2-tuple as part of the routines interface. The
first element is the ``larpix.larpix.Controller`` object containing
up-to-date configurations, IO, and Logger objects. The second element is
the "return value" or result of the routine, which must be
JSON-encodable (numbers, strings, booleans, or ``None`` literals, or dicts or lists
of those literals).

Writing routines
----------------

When new routines are being written, they can be tested using the
``test_routine`` convenience method. A test session might look like

>>> from larpixdaq.routines import init_routines, test_routine
>>> init_routines('.')  # assuming you have routines in the current directory
>>> from larpix.larpix import Controller
>>> controller = Controller()
>>>  # ... then set up the controller
>>> test_routine('my_routine', controller, print, print, (arg1, arg2, arg3))
(<larpix.larpix.Controller at 0x7f3bf6d5c8d0>, <routine return value>)

Sharing and finding routines
----------------------------

A global repository of LArPix DAQ routines is being set up at
<https://github.com/larpix/larpix-daq-routines>. Contact the LBNL LArPix
contacts for help, and/or file an issue or pull request.
"""
import importlib
import os
import sys

ROUTINES = {}
_routine_files = {}

def init_routines(location=None):
    """Collect and register routines.

    :param location: The directory to look for routines files.
        (optional. If absent or ``None`` then look inside the
        ``larpixdaq.routines`` package directory.)
    """
    if location is None:
        location = os.path.dirname(__file__)
    if location not in sys.path:
        sys.path.insert(0, location)
    importlib.invalidate_caches()
    routine_files = [os.path.splitext(x)[0] for x in os.listdir(location) if (
        x.endswith('.py') and x != __file__)]
    for routine_file in routine_files:
        if routine_file in _routine_files:
            module_old = _routine_files[routine_file]
            module = importlib.reload(module_old)
        else:
            module = importlib.import_module(routine_file)
        _routine_files[routine_file] = module
        if hasattr(module, 'registration'):
            ROUTINES.update(module.registration)
    print(ROUTINES)

def test_routine(name, controller, send_data=None, send_info=None,
        args=()):
    """Run the given routine in a test or custom environment.

    This is useful during development, where a routine can be executed
    without loading up the entire DAQ system.

    A reasonable way to mock up ``send_data`` and ``send_info`` is
    simply to pass the ``print`` function (in Python 3; in Python 2
    first ``from __future__ import print_function``).

    :param name: the name of the routine to test
    :param controller: the ``larpix.larpix.Controller`` object
    :param send_data: a function to call to send data down the pipeline.
        Signature: ``def send_data(packets_to_send)``.  (optional, default or
        ``None`` results in ``def send_data(to_send): return None``)
    :param send_info: a function to call to send info messages down the
        pipeline. Signature: ``def send_info(str_to_send)``. (optional,
        default or ``None`` results in ``def send_data(to_send): return
        None``)
    :param args: the arguments to pass to the routine, as an iterable.
        If there's only one argument, it's recommended to use a list
        (``[arg1]``) since a 1-tuple requires a comma which is easy to
        forget (``(arg1,)``). (optional, default: ``()``)
    :returns: the return value of the routine
    """
    if send_data is None:
        def send_data(to_send):
            return None
    if send_info is None:
        def send_info(to_send):
            return None
    routine = ROUTINES[name]
    func = routine.func
    return func(controller, send_data, send_info, *args)

class Routine(object):
    '''
    Represents a routine run using the LArPix Control board.

    Routine functions are passed three arguments for the purpose of
    interacting with the LArPix-control and DAQ systems, so they should
    have the following signature:

    ```
    def func(controller, send_data, send_info, [...])
    ```

    - ``controller`` is the current instance of the LArPix-control
      Controller object
    - ``send_data`` is a function which sends data to the DAQ pipeline.
      The first argument is the data, in bytes. The second is optional
      and is a dict with metadata. Default metadata of Unix timestamp
      and component name is automatically added to the supplied dict.
    - ``send_info`` is a function which sends an informational message
      to the DAQ pipeline. The first and only argument is the message as
      a string.
    - Any other arguments must be listed in order in self.params.

    '''
    def __init__(self, name, func, params=None):
        if params is None:
            params = []
        self.name = name
        self.func = func
        self.params = params




'''
The operator interface for the LArPix DAQ system.

'''
import xylem

end_receive_loop_headers = {
        'ACTIONS',
        'ACTION RESULT',
        'STATE',
        'STATE REQUEST',
        'CLIENTS',
        }

class Operator(object):
    """The LArPix DAQ Operator module provides the interface into the
    DAQ core for all DAQ operations.

    The Operator class handles all of the needs of an DAQ operator:
    start/end runs, load/validate configurations, run calibrations,
    examine data samples and data rates, etc.

    Operator methods interact with the DAQ Core to accomplish the
    desired behavior. The DAQ Core can send
    multiple responses for a single request - e.g. an immediate
    acknowledgement of receipt and then the eventual result. The
    methods implementing these interactions return generator
    iterators <https://docs.python.org/3/glossary.html#term-generator>
    rather than values. The way to call these functions usually looks
    like::

        o = Operator()
        final_responses = []
        for response in o.run_routine('example'):
            print(response)
            # interact with response object within loop
        # When the loop ends, the last response received is still saved in
        # the response object
        final_responses.append(response)

    Each method has an optional keyword parameter ``timeout``. If
    omitted (or if ``None``), then the Operator will wait forever for a
    response from the DAQ Core. If ``timeout`` is provided, then the
    Operator will yield ``None`` as an indication of an error after a
    maximum wait of 10 times the given timeout, in seconds.

    .. note:: if the timeout limit is reached and the "missing" message
        arrives later, it will be confused with future results.
        Operator objects do not maintain useful internal state, so it is
        acceptable (and recommended!) to initialize a new Operator if
        the current Operator ran into a timeout issue.

    :param address: the TCP address of the DAQ Core. The port will be
        added automatically. (Optional, if omitted or ``None``, will
        default to ``tcp://127.0.0.1``.)
    """

    def __init__(self, address=None):
        if address is None:
            address = 'tcp://127.0.0.1:5551'
        else:
            address += ':5551'
        self._controller = xylem.Controller(address)

    def cleanup(self):
        self._controller.cleanup()

    def _receive_loop(self, timeout=None):
        header = None
        max_loops = 10
        nloops = 0
        while (header not in end_receive_loop_headers
                and nloops < max_loops):
            result = self._controller.receive(timeout)
            nloops += 1
            if result is not None:
                header = result['header']
                yield result
        if result is None:
            yield None

    def get_boards(self, timeout=None):
        self._controller.send_action('LArPix board', 'get_boards', [])
        for result in self._receive_loop(timeout):
            yield result

    def load_board(self, filename, timeout=None):
        self._controller.send_action('LArPix board', 'load_board',
                [filename])
        for result in self._receive_loop(timeout):
            yield result

    def retrieve_pixel_layout(self, filename, timeout=None):
        self._controller.send_action('Online monitor', 'retrieve_pixel_layout',
                [filename])
        for result in self._receive_loop(timeout):
            yield result

    def load_pixel_layout(self, filename, timeout=None):
        self._controller.send_action('Online monitor', 'load_pixel_layout',
                [filename])
        for result in self._receive_loop(timeout):
            yield result

    ### Configurations

    def write_configuration(self, chip, timeout=None):
        '''
        Send the configuration values from software onto the ASIC.

        '''
        self._controller.send_action('LArPix board', 'write_config',
                [chip])
        for result in self._receive_loop(timeout):
            yield result

    def read_configuration(self, chip, timeout=None):
        '''
        Read the configuration values from the ASIC.

        '''
        self._controller.send_action('LArPix board', 'read_config',
                [chip])
        for result in self._receive_loop(timeout):
            yield result

    def validate_configuration(self, chip, timeout=None):
        '''
        Read the configuration from the specified LArPix ASIC and return
        ``(True/False, {name: (actual, stored)})``.

        '''
        self._controller.send_action('LArPix board', 'validate_config',
                [chip])
        for result in self._receive_loop(timeout):
            yield result

    def retrieve_configuration(self, chip, timeout=None):
        '''
        Return a dict of the current configuration stored in software
        for the given chipid.

        '''
        self._controller.send_action('LArPix board', 'retrieve_config',
                [chip])
        for result in self._receive_loop():
            yield result

    def send_configuration(self, updates, timeout=None):
        '''
        Send the given configuration updates to the LArPix control
        software.

        Updates should be a dict with keyed by chip ID.

        '''
        self._controller.send_action('LArPix board', 'send_config',
                [updates])
        for result in self._receive_loop():
            yield result


    ### Calibrations

    def list_routines(self):
        '''
        Return a list of routines/calibrations.

        '''
        self._controller.send_action('LArPix board',
                'list_routines', [])
        for result in self._receive_loop():
            yield result

    def run_routine(self, name, *args, timeout=None):
        '''
        Run the given routine and return the routine's output.

        '''
        self._controller.send_action('LArPix board',
                'run_routine', [name] + list(args))
        for result in self._receive_loop(timeout):
            yield result


    ### Physics runs

    def prepare_physics_run(self, timeout=None):
        '''
        Enter the "READY" DAQ state so that all DAQ components are ready
        to begin the physics run.

        '''
        self._controller.request_state_change('READY')
        for result in self._receive_loop(timeout):
            yield result

    def begin_physics_run(self, timeout=None):
        '''
        Begin taking physics data, activate online data monitoring and
        analytics, and store the data in offline storage.

        '''
        self._controller.request_state_change('RUN')
        for result in self._receive_loop(timeout):
            yield result


    def end_physics_run(self, timeout=None):
        '''
        Stop taking physics data, deactivate online data monitoring and
        analytics, and finalize the offline storage.

        '''
        self._controller.request_state_change('STOP')
        for result in self._receive_loop(timeout):
            yield result




from larpixdaq.routines import Routine

def _hello_world(controller, send_data, send_info, *args):
    to_return = "Hello, %s!" % args
    send_info(to_return)
    return controller, to_return

hello_world = Routine('hello_world', _hello_world, ['name'])

registration = {'hello_world': hello_world}



import logging
import argparse
import time

from xylem import Aggregator
from larpix.larpix import Controller, Packet

from larpixdaq.packetformat import fromBytes

class LArPixAggregator(object):
    """The data aggregator for LArPix.

    The aggregator connects the producer (LArPix board) to any number of
    data consumers such as data monitor and offline storage.

    In the future, the aggregator will support receiving data from
    multiple LArPix boards.

    Example invocation::

        python -m larpixdaq.aggregator tcp://127.0.0.1:5002

    :var aggregator: the xylem Aggregator object
    :var state: the DAQ state of the xylem Aggregator component

    :param output_address: the full TCP address (including port number)
        that data will be published to
    :param core_address: the full TCP address (including port number) of
        the DAQ core
    """

    def __init__(self, output_address, core_address):
        kwargs = {
                'core_address': core_address,
                'heartbeat_time_ms': 300,
                }
        self.aggregator = Aggregator(output_address, name='LArPix aggregator',
                connections=['BOARD'], **kwargs)
        self.aggregator.request_state()
        self.state = self.aggregator.state

    def run(self):
        """Initiate the event loop of reading and rebroadcasting
        messages and data.
        """
        while True:
            messages = self.aggregator.receive(1)
            for message in messages:
                if message[0] == 'DATA':
                    _, metadata, data = message
                    self.aggregator.broadcast(metadata, data)
                elif message[0] == 'INFO':
                    self.aggregator.log('DEBUG', 'Rebroadcasting (%s, %s)' %
                            (message[1], message[2]))
                    self.aggregator.rebroadcast_info(message[1], message[2])
            if self.state != self.aggregator.state:
                self.aggregator.log('DEBUG', 'State update. New state: %s' % aggregator.state)
                self.state = self.aggregator.state

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Launch the data '
            'aggregator for LArPix DAQ')
    parser.add_argument('address',
            help='The address to publish data to including port number')
    parser.add_argument('--core', default='tcp://127.0.0.1',
            help='The address of the DAQ Core, not including port number')
    parser.add_argument('-d', '--debug', action='store_true',
            help='Enter debug (verbose) mode')
    args = parser.parse_args()
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    aggregator = LArPixAggregator(args.address, args.core + ':5551')
    try:
        aggregator.run()
    except KeyboardInterrupt:
        pass
    finally:
        aggregator.aggregator.cleanup()



from __future__ import absolute_import
from __future__ import print_function
import argparse
import logging
import time
import ast
import json
import random

from xylem import Producer
import larpix.quickstart as larpix_quickstart
from larpix.io.fakeio import FakeIO
from larpix.io.zmq_io import ZMQ_IO
from larpix.io.multizmq_io import MultiZMQ_IO
import larpix.configs as configs
import larpix.larpix as larpix

from larpixdaq.packetformat import toBytes
from larpixdaq.routines import ROUTINES, init_routines
from larpixdaq.logger_producer import DAQLogger

class LArPixProducer(object):
    """The entry point of LArPix data into the xylem DAQ pipeline.

    On initialization, the LArPixProducer object will be connected to
    the DAQ Core, loaded up with the ``'pcb-1'`` Controller
    configuration, connected to the appropriate IO instance, and set up
    with a ``larpix.logger.Logger`` subclass instance that logs all data
    into xylem (disabled on initialization, accessible at
    ``self.board.logger``).

    The producer is the component with direct contact into the LArPix
    environment. As such, the producer receives data from the data board
    and sends it into the DAQ chain. It also sends configuration
    commands to the LArPix ASICs and runs custom DAQ routines such as
    threshold scans and calibrations.

    Implementation-wise, this all happens via a larpix-control
    Controller object (not to be confused with the DAQ Controller).

    Custom routines can be implemented using the
    :py:mod:`larpixdaq.routines` package. Custom
    Routines are managed in a Routine object, in which you should store
    the routine name, function handle/reference, and list of parameters.
    (TODO!!! allow for documentation for custom routines.) Routines can
    access the DAQ functionality via their arguments controller,
    send_data, send_info. They can also accept additional arguments.
    Routines must return a tuple of (controller, result) where result is
    the output of the routine (e.g. a list of thresholds, or even simply
    the string "success") which must be JSON-serializable.

    :var producer: the xylem Producer object used to send data
    :var board: the ``larpix.larpix.Controller`` instance used to gather
        data
    :var current_boardname: the short name of the layout/configuration
        used for the Controller object, e.g. ``'pcb-1'``.
    :var state: the DAQ State of the xylem Producer component

    :param output_address: the full TCP address (including port number)
        that data will be published to
    :param core_address: the full TCP address (including port number) of
        the DAQ Core
    :param use_fakeio: ``True`` to use a ``FakeIO`` object rather than
        real data
    :param config_file: the file path of the IO configuration file to
        use with a ``ZMQ_IO`` object (unused if ``use_fakeio``)
    """

    def __init__(self, output_address, core_address, io_config):
        kwargs = {
                'core_address': core_address,
                'heartbeat_time_ms': 300,
        }
        self.producer = Producer(output_address, name='LArPix board', group='BOARD', **kwargs)
        self.board = larpix.Controller()
        io_class = io_config[0]
        if io_class == 'FakeIO':
            self.board.io = FakeIO()
        elif io_class == 'ZMQ_IO':
            config_file = io_config[1]
            self.board.io = ZMQ_IO(config_file)
        elif io_class == 'MultiZMQ_IO':
            config_file = io_config[1]
            self.board.io = MultiZMQ_IO(config_file)
        else:
            raise ValueError('Invalid IO class from --io-config: %s' %
                    io_config[0])
        self.board.load('controller/pcb-1_chip_info.json')
        self.current_boardname = 'pcb-1'
        self.board.logger = DAQLogger(self.producer)
        self.state = ''
        run = False
        configurations = {
                'startup': 'startup.json',
                'quiet': 'quiet.json',
        }
        ############
        # Routines #
        ############
        init_routines()
        self.producer.register_action(*self._get_register_action_args('write_config'))
        self.producer.register_action(*self._get_register_action_args('read_config'))
        self.producer.register_action(*self._get_register_action_args('validate_config'))
        self.producer.register_action(*self._get_register_action_args('retrieve_config'))
        self.producer.register_action(*self._get_register_action_args('send_config'))
        self.producer.register_action(*self._get_register_action_args('get_boards'))
        self.producer.register_action(*self._get_register_action_args('load_board'))
        self.producer.register_action(*self._get_register_action_args('list_routines'))
        self.producer.register_action(*self._get_register_action_args('run_routine'))
        self.producer.register_action(*self._get_register_action_args('sleep'))
        self.producer.request_state()

    def _get_register_action_args(self, name):
        """Return a tuple that can be passed as
        ``*self._get_register_action_args(name)`` to
        producer.register_action.

        :param name: the (string) name of the method to use for the
            routine. Doubles as the routine name.
        """
        method = getattr(self, name)
        return (name, method, method.__doc__)

    def write_config(self, key, registers_str=''):
        """Send the given configuration to the board.

        :param key: the chip key to send the configuration to
        :param registers_str: the configuration registers to send,
            specified as an int, a list of ints, or a string specifying a
            literal int or list of ints (e.g. ``'[1, 2, 10]'``).
        """
        if registers_str:  # treat as int or list
            registers = ast.literal_eval(registers_str)
        else:
            registers = None
        self.board.write_configuration(key, registers, 0, None)
        return 'success'

    def read_config(self, key, registers_str=''):
        """Read configurations from the board.

        :param key: the chip key to send the configuration to
        :param registers_str: the configuration registers to send,
            specified as an int, a list of ints, or a string specifying a
            literal int or list of ints (e.g. ``'[1, 2, 10]'``).
        """
        if registers_str:  # treat as int or list
            registers = ast.literal_eval(registers_str)
        else:
            registers = None
        if isinstance(self.board.io, FakeIO):
            chip = self.board.get_chip(key)
            packets = chip.get_configuration_packets(
                    larpix.Packet.CONFIG_WRITE_PACKET)
            for p in packets:
                p.packet_type = larpix.Packet.CONFIG_READ_PACKET
                p.assign_parity()
            self.board.io.queue.append((packets, b'some bytes'))
        self.board.read_configuration(key, registers, message=None)
        packets = self.board.reads[-1]
        result = '\n'.join(str(p) for p in packets if p.packet_type
                == p.CONFIG_READ_PACKET)
        return result

    def validate_config(self, key):
        """Read configurations from the board and compare to those stored
        in software, returning ``True`` if they're equal.

        :param key: the chip key whose configuration will be validated
        """
        if isinstance(self.board.io, FakeIO):
            chip = self.board.get_chip(key)
            packets = chip.get_configuration_packets(
                    larpix.Packet.CONFIG_WRITE_PACKET)
            for p in packets:
                p.packet_type = larpix.Packet.CONFIG_READ_PACKET
            self.board.io.queue.append((packets, b'some bytes'))
        result = self.board.verify_configuration(key)
        return result

    def retrieve_config(self, key):
        """Return the current configuration stored in software for the
        given chip.

        :param key: the chip key whose configuration will be retrieved
        """
        chip = self.board.get_chip(key)
        return chip.config.to_dict()

    def send_config(self, updates):
        """Apply the given updates to the software configuration.

        :param updates: a dict of configuration register updates
        compatible with ``larpix.larpix.Config.from_dict``.
        """
        for key, chip_updates in updates.items():
            chip = self.board.get_chip(key)
            chip.config.from_dict(chip_updates)
        return 'success'

    def get_boards(self):
        """List the available boards and chip keys, and the current board
        name.
        """
        boards = [
                'pcb-1',
                'pcb-2',
                'pcb-3',
                'pcb-4',
                'pcb-5',
                'pcb-6',
                'pcb-10',
                ]
        board_data = []
        for boardname in boards:
            result = configs.load('controller/' + boardname +
                    '_chip_info.json')
            boarditem = {
                    'name': result['name'],
                    'chips': result['chip_list'],
                    }
            board_data.append(boarditem)
        return {'data': board_data, 'current': self.current_boardname}

    def load_board(self, filename):
        """Load the board (Controller) configuration located at the given
        filename.

        :param filename: the file name to load. If using a pre-installed
            configuration, the file name must begin with the standard
            ``'controller/'`` directory prefix.
        """
        data = configs.load(filename)
        self.current_boardname = data['name']
        self.board.load(filename)
        return 'success'

    @staticmethod
    def list_routines():
        """List the available routines."""
        init_routines()
        return [{
            'name': name,
            'params': [{'name': p, 'type': 'input'} for p in
                r.params],
            }
            for name, r in ROUTINES.items()
            ]

    def run_routine(self, name, *args):
        """Run the given routine.

        :param name: the name of the routine to run
        :param args: all subsequent arguments are passed in order to the
            routine as parameters
        """
        def send_data(packet_list, metadata=None):
            self.producer.produce(toBytes(packet_list), metadata)
            return
        self.board, result = ROUTINES[name].func(self.board, send_data,
                self.producer.send_info, *args)
        return result

    @staticmethod
    def sleep(time_in_sec):
        """Sleep and return success."""
        delay = int(time_in_sec)
        time.sleep(delay)
        return 'success'

    def run(self):
        """Event loop of checking for DAQ commands, checking for new
        data, and repeating.

        If the DAQ State is RUN, the data logger will be enabled so that
        newly-arrived data will be sent to the xylem pipeline. In all
        other DAQ States, the data logger will be disabled so data will
        not be send down the pipeline.

        If the IO object on ``self.board`` is a FakeIO object, fake data
        will be generated to mimic data arriving from the LArPix board.
        """
        while True:
            self.producer.receive(0.25)
            if self.state != self.producer.state:
                old_state = self.state
                new_state = self.producer.state
                print('State update: New state: %s' % new_state)
                if old_state == 'RUN':
                    self.producer.send_info('Ending run')
                    self.board.logger.disable()
                if new_state == 'RUN':
                    self.producer.send_info('Beginning run')
                    self.board.logger.enable()
                    fake_timestamp = 0
                self.state = self.producer.state
            if self.state == 'RUN':
                if not self.board.io.is_listening:
                    logging.debug('about to start listening')
                    self.board.start_listening()
                if isinstance(self.board.io, FakeIO):
                    packets = []
                    for _ in range(300):
                        p = larpix.Packet()
                        p.timestamp = fake_timestamp % 16777216
                        p.dataword = int(sum(random.random() for _ in range(256)))
                        chip = random.choice(list(self.board.chips.values()))
                        p.chipid = chip.chip_id
                        p.channel_id = random.randint(0, 31)
                        fake_timestamp += 1
                        p.assign_parity()
                        p.chip_key = '%d-%d-%d' % (1, 1, chip.chip_id)
                        packets.append(p)
                    self.board.io.queue.append((packets, p.bytes() + b'\x00'))
                data = self.board.read()
            else:
                if self.board.io.is_listening:
                    self.board.stop_listening()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Launch the data '
        'interface between LArPix and the xylem DAQ pipeline')
    parser.add_argument('address',
            help='The address to publish data to including port number')
    parser.add_argument('--core', default='tcp://127.0.0.1',
            help='The address of the DAQ Core, not including port number')
    parser.add_argument('-d', '--debug', action='store_true',
            help='Enter debug (verbose) mode')
    parser.add_argument('--io-config', nargs='+', required=True,
            help='<IO class> [<IO config file>], e.g. "ZMQ_IO io/default.json"')
    args = parser.parse_args()
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    address = args.address
    core_url = args.core
    core_address = core_url + ':5551'
    producer = LArPixProducer(address, core_address, args.io_config)
    try:
        producer.run()
    except KeyboardInterrupt:
        pass
    finally:
        producer.producer.cleanup()



from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md')) as f:
    long_description = f.read()

with open(os.path.join(here, 'VERSION')) as f:
    version = f.read()


setup(
        name='larpix-daq',
        version=version,
        description='LArPix DAQ system',
        long_description=long_description,
        long_description_content_type='text/markdown',
        author='Sam Kohn, LBNL Neutrino Group',
        author_email='skohn@lbl.gov',
        keywords='dune larpix',
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python',
            'Topic :: Scientific/Engineering :: Physics',
            'Topic :: Scientific/Engineering',
            ],
        packages=find_packages(),
        install_requires=['xylem-daq ~= 0.3.1', 'flask >=1.0.0', 'flask-socketio >=3.0',
            'eventlet >= 0.24', 'requests ~= 2.18',
            'larpix-control ~= 2.3.0'],
)



'''
Formatter for LArPix data packets + medatadata to/from bytestreams.

'''
import json
import struct

from larpix.larpix import Packet, TimestampPacket

def get_packet(packet_bytes):
    byte_marker = packet_bytes[0]
    if byte_marker == 0:
        p = Packet(packet_bytes[1:8])
        p.chip_key = '%d-%d-%d' % (
                packet_bytes[8],
                packet_bytes[9],
                p.chipid)
        return p
    elif byte_marker == 1:
        return TimestampPacket(code=packet_bytes[1:8])

def format_packet(packet):
    if isinstance(packet, Packet):
        result = b'\x00' + packet.bytes()
        if packet.chip_key is None:
            io_group = 0
            io_channel = 0
        else:
            io_group = packet.chip_key.io_group
            io_channel = packet.chip_key.io_channel
        result += struct.pack('BB', io_group, io_channel)
        return result
    elif isinstance(packet, TimestampPacket):
        return b'\x01' + packet.bytes() + b'\x00\x00'


def toBytes(packets):
    '''
    Format the packets into a single bytestream.

    Format:

    - First 2 bytes, interpreted as integer values, are the version (major
    is byte 0, minor is byte 1).
    - Next is the ``b'/'`` character (`` == b'\x2F'``)
    - Then each packet takes 10 bytes:
        - First the packet-type indicator (``b'\x00'`` for data packets,
          ``b'\x01'`` for timestamps, etc.)
        - Then the packet data itself:
            - For data packets, 7 UART bytes from ``Packet.bytes()``
              (bytes 1-7), then the chip key data in 2 bytes: byte 8 is the
              IO group, and byte 9 is the IO channel. (The chip ID is
              contained in the UART bytes.)
            - For the timestamp packets, 7 timestamp bytes followed by
              2 unused bytes.

    '''

    return b'\x00\x01/' + b''.join(format_packet(p) for p in packets)

def fromBytes(bytestream):
    version, slash, stream = bytestream.partition(b'/')
    if slash == b'':
        raise ValueError('No version found')
    split = [stream[n:n+10] for n in range(0, len(stream), 10)]
    packets = [get_packet(x) for x in split]
    return packets

def to_unicode_coding(packets):
    '''
    Return an encoding of the packet bytes as a Unicode string.

    '''
    return json.dumps([p.export() for p in packets])

def from_unicode_coding(unicode_stream):
    '''
    Convert the Unicode stream back into Packets.

    '''
    raise NotImplementedError()

def toDict(packets):
    return [p.export() for p in packets]

def fromDict(jsonlist):
    type_map = {
            'data': 0,
            'test': 1,
            'config write': 2,
            'config read': 3,
            }
    packets = []
    for item in jsonlist:
        p = Packet()
        p.packet_type = item['type']
        p.chipid = item['chipid']
        p.parity = item['parity']
        if item['type'] == 'data':
            p.channel = item['channel']
            p.timestamp = item['timestamp']
            p.dataword = item['adc_counts']
            p.fifo_half_flag = item['fifo_half']
            p.fifo_full_flag = item['fifo_full']
        elif item['type'] == 'test':
            p.counter = item['counter']
        elif (item['type'] == 'config write'
                or item['type'] == 'config read'):
            p.register_address = item['register']
            p.register_data = item['value']
    return packets




from __future__ import absolute_import
from __future__ import print_function
from xylem import Core
from xylem.EventHandler import EventHandler
import os
import argparse
import requests

class LArPixCore(object):
    """The core is responsible for managing and monitoring the DAQ
    system components. This includes tracking which components are
    connected, broadcasting the DAQ state, and sending operator commands
    to different components.

    The core must be the first script launched, since other components
    will exit if they cannot connect to the core.

    Operator/user interaction with the DAQ Core, and therefore the
    rest of the DAQ system, happens exclusively through the LArPix DAQ
    Operator module. See the :py:class:`Operator documentation
    <.Operator>` for available commands.

    """
    def __init__(self, address, log_address):
        self.core = Core(address, log_address)
        self._allowed_states = ['READY', 'RUN', 'STOP']
        self.core.state = 'STOP'
        self.core.isStateAllowed = lambda x: x in self._allowed_states
        server = 'http://localhost:5000/'

        def announce_new_client(client_name, all_client_names):
            try:
                r = requests.post(server + '/component',
                        json={'new':client_name, 'all':all_client_names})
            except:
                pass
        def announce_lost_client(client_name, all_client_names):
            try:
                r = requests.delete(server + '/component',
                        json={'lost':client_name, 'all':all_client_names})
            except:
                pass
        def announce_state_change(new_state, old_state):
            try:
                r = requests.post(server + '/state', json={'new':
                    new_state, 'old': old_state})
            except:
                pass
        announce = EventHandler('new_component', announce_new_client)
        announce_lost = EventHandler('lost_component', announce_lost_client)
        announce_state = EventHandler('state_change', announce_state_change)
        self.core.addHandler(announce)
        self.core.addHandler(announce_lost)
        self.core.addHandler(announce_state)

    def run(self):
        """Enter the event loop for the Core."""
        self.core.run()

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--address', default='tcp://127.0.0.1',
            help='base address for ZMQ connections')
    parser.add_argument('--log-address', default='tcp://127.0.0.1:5678',
            help='Address to connect to global log')

    args = parser.parse_args()
    base_address = args.address + ':'

    core = LArPixCore(base_address + '5551', args.log_address)

    try:
        core.run()
    except KeyboardInterrupt:
        pass



"""The LArPix DAQ System records data from LArPix.

LArPix DAQ consists of a set of executable back-end modules, plus an
operator API. The system is built on top of the xylem DAQ framework. The
documentation here is intented to be self-contained.

Xylem allows for a networked DAQ system, where different components live
on different computers and send data and commands over the local network
and/or the Internet. This documentation will assume that all components
are being executed on the same machine, which will be referred to
by the loopback address 127.0.0.1. If you want to use the networked
functionality, simply replace 127.0.0.1 with the IP address of the
machine that the script is running on. Additionally, all scripts take
optional ``--core`` and ``--log-address`` arguments which default to ``tcp://127.0.0.1`` and
so will not be used in the documentation below. So you will need to pass
an explicit ``--core`` and/or ``--log-address`` argument with the IP address of the DAQ Core (if
it's not running on the same machine as the script), such as ``--core
tcp://labcomputer.university.edu``.

The suggested way to run all the scripts on one machine is to use
multiple terminals rather than sending scripts to the background, at
least until the software becomes stable.

Startup sequence
----------------

The first step is to launch the DAQ Log, which is a plain xylem module::

    python -m xylem.Log -p tcp://127.0.0.1:5678 -o log.txt

The ``-p`` argument lists the port to listen on (5678 is a convention
but not required), and ``-o`` give sthe location of the output file.
This Log script will print INFO-level messages and higher to stdout and
DEBUG-level messages and higher (i.e. all messages) to the output file.
To get all messages on stdout, simply start the log, then run ``tail
-f log.txt``, which will print each line of log.txt to stdout as it
arrives.

The DAQ Log is used for debugging and accountability purposes. It is not
intended to be particularly useful in day-to-day (production) operation.

The first "real" LArPix DAQ component to start is the DAQ Core, which
monitors and coordinates the other components::

    python -m larpixdaq.core

The rest of the DAQ components can be started in any order, but the
suggested order is the order of data flow::

    python -m larpixdaq.producer tcp://127.0.0.1:5001 --io-config FakeIO
    python -m larpixdaq.aggregator tcp://127.0.0.1:5002
    python -m larpixdaq.offline_storage
    python -m larpixdaq.online_monitor

Interacting with the DAQ
------------------------

All of these scripts are *not* interactive. All interactivity is
performed via an :py:class:`~operator.Operator` object, which can be scripted or
used in an interactive Python session as follows::

    from larpixdaq.operator import Operator
    op = Operator(core_IP_address)
    for response in op.prepare_physics_run():
        # do something with each response from the Core
        print(response)
    # do something with the final result, still saved as response
    print("Once again, the final result is:\\n", response)

The LArPix Webapp server uses this Operator API, so if you are using the
webapp you do not need to start your own Operator.

About port numbers
------------------

The command-line arguments for the DAQ scripts require you to specify
the port number for each component. By convention, the port numbers
should start at 5001 for the first component and then increment from
there. Only components which *send* data to other components need a
port number and IP address, which is why the offline storage and online
monitor don't get one.
"""
